'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Licensed Materials - Property of IBM
 * IBM Cognos Products: BI Cloud (C) Copyright IBM Corp. 2020, 2024
 * US Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
define(['dashboard-analytics/util/DashboardFormatter', './Column', './Decorated'], function (DashboardFormatter, Column, Decorated) {
	/** Used to replace missing categorical values .
  * As categorical columns usally contain letters, it is 
  * unlikely to have "-99" in it. 
  */
	var CAT_MISSING_VALUE = '-99';

	/**
  * The what-if extensions takes a query result, which would normally be drawn in a visualization, as input data.
  * This means that the query result, to this extensions, is source data. An AI model can be trained upon it and
  * that model can be used to predict values. These predictions may be given to a visualization, which requires
  * a DataSet too. This is why this module encodes two types of data sets: a source table and visualization data.
  */

	/**
  * Check whether a measure value should be considered "missing", 
  * e.g. it is not a valid number.
  * @param {unknown} _value The value to test.
  * @returns {boolean} Whether the value is considered missing.
  */
	function isMissingMeasureValue(_value) {
		return _value === undefined || _value === null || isNaN(_value);
	}

	/**
  * Check whether a categorical value should be considered "missing"
  * @param {unknown} _value The value to test.
  * @returns {boolean} Whether the value is considered missing.
  */
	function isMissingCatValue(_value) {
		return _value === undefined || _value === null;
	}

	function getMeanForMeasuredInputCols(sourceDataPoints, columns, resultItemList) {
		var total = new Array(columns.length).fill(0);
		var counts = new Array(columns.length).fill(0);

		for (var rowIndex = 0; rowIndex < sourceDataPoints.length; rowIndex++) {
			for (var columnIndex = 0; columnIndex < columns.length; columnIndex++) {
				var column = columns[columnIndex];
				if (column.type === 'measure' && !Column.isTargetColumn(column)) {
					// Note that for every non aggregated column, value will be a tuple index.
					var value = sourceDataPoints[rowIndex].pt[columnIndex];
					// Reading the value gets a bit odd; due to the aggregate none measures are actually reported
					// as tuples, for any column but the target column.
					var specObj = void 0;
					if ((typeof value === 'undefined' ? 'undefined' : _typeof(value)) !== 'object') {
						specObj = resultItemList[columnIndex].getValue(value)[0];
					}
					value = typeof specObj !== 'undefined' ? specObj.d : value.v;
					if (!isMissingMeasureValue(value)) {
						total[columnIndex] += value;
						counts[columnIndex]++;
					}
				}
			}
		}
		for (var _columnIndex2 = 0; _columnIndex2 < columns.length; _columnIndex2++) {
			if (counts[_columnIndex2] > 0) {
				// compute means 
				total[_columnIndex2] = total[_columnIndex2] / counts[_columnIndex2];
			}
		}
		return total;
	}

	function createSourceTable(sourceDataPoints, columns, resultItemList) {
		// The data is stored in a flattened (1 dimensional) Array, which is a memory optimization.
		// Initializing this array with the expected length prevents the need to grow the Array's backing storage.
		var rowData = new Array(sourceDataPoints.length * columns.length);
		/* 
   * How missing values are handled:
   * for INPUT measure columns, we replace missing values with col mean
   * for the target column, we ignore any row with a missing value 
   * for categorical columns, missing values are replaced by a new category CAT_MISSING_VALUE
  */
		var colMean = getMeanForMeasuredInputCols(sourceDataPoints, columns, resultItemList);
		var insertRowIndex = 0;
		for (var rowIndex = 0; rowIndex < sourceDataPoints.length; rowIndex++) {
			// Track whether the row should be included.
			// Since AI models cannot handle missing values this is primarily based on whether the row has missing values.
			var includeRow = true;
			// Do not stop processing a row if includeRow turns to false; Tuples may still be discovered.
			for (var columnIndex = 0; columnIndex < columns.length; columnIndex++) {
				var column = columns[columnIndex];
				// Note that for every non aggregated column, value will be a tuple index.
				var value = sourceDataPoints[rowIndex].pt[columnIndex];
				if (!Column.isContextColumn(column)) {
					// Reading the value gets a bit odd; due to the aggregate none measures are actually reported
					// as tuples, for any column but the target column.
					var specObj = void 0;
					if ((typeof value === 'undefined' ? 'undefined' : _typeof(value)) !== 'object') {
						specObj = resultItemList[columnIndex].getValue(value)[0];
					}
					value = typeof specObj !== 'undefined' ? specObj.d : value.v;
					if (column.type === 'measure') {
						if (isMissingMeasureValue(value)) {
							if (Column.isTargetColumn(column)) {
								value = null; // normalize missing values to null
								includeRow = false;
							} else {
								value = colMean[columnIndex];
							}
						}
					} else {
						// column.type === 'categorical'
						if (isMissingCatValue(value)) {
							value = CAT_MISSING_VALUE;
						}
					}
				}
				// for context columns: use tuple index as value
				rowData[insertRowIndex * columns.length + columnIndex] = value;
			}
			if (includeRow) {
				// only count the row if it is included.
				insertRowIndex++;
			}
		}
		// In case any rows where filtered out: trim the row data to the correct length.
		rowData.length = insertRowIndex * columns.length;
		return new SourceTable(columns, rowData);
	}

	function getMeanForMeasuredInputColsV2(sourceData, columns) {
		var total = new Array(columns.length).fill(0);
		var counts = new Array(columns.length).fill(0);

		for (var rowIndex = 0; rowIndex < sourceData.getRowCount(); rowIndex++) {
			for (var columnIndex = 0; columnIndex < columns.length; columnIndex++) {
				var column = columns[columnIndex];
				if (column.type === 'measure' && !Column.isTargetColumn(column)) {
					var value = sourceData.getValue(rowIndex, columnIndex);
					// Reading the value gets a bit odd; due to the aggregate none measures are actually reported
					// as tuples, for any column but the target column.
					var specObj = void 0;
					if (column.type !== 'measure') {
						specObj = value[0];
					}
					value = typeof specObj !== 'undefined' ? specObj.d : Array.isArray(value) ? value[0].v : value.v;
					if (!isMissingMeasureValue(value)) {
						total[columnIndex] += value;
						counts[columnIndex]++;
					}
				}
			}
		}
		for (var _columnIndex3 = 0; _columnIndex3 < columns.length; _columnIndex3++) {
			if (counts[_columnIndex3] > 0) {
				// compute means 
				total[_columnIndex3] = total[_columnIndex3] / counts[_columnIndex3];
			}
		}
		return total;
	}

	function createSourceTableV2(sourceData, columns) {
		// The data is stored in a flattened (1 dimensional) Array, which is a memory optimization.
		// Initializing this array with the expected length prevents the need to grow the Array's backing storage.
		var rowData = new Array(sourceData.getRowCount() * columns.length);
		/* 
   * How missing values are handled:
   * for INPUT measure columns, we replace missing values with col mean
   * for the target column, we ignore any row with a missing value 
   * for categorical columns, missing values are replaced by a new category CAT_MISSING_VALUE
  */
		var colMean = getMeanForMeasuredInputColsV2(sourceData, columns);
		var insertRowIndex = 0;
		for (var rowIndex = 0; rowIndex < sourceData.getRowCount(); rowIndex++) {
			// Track whether the row should be included.
			// Since AI models cannot handle missing values this is primarily based on whether the row has missing values.
			var includeRow = true;
			// Do not stop processing a row if includeRow turns to false; Tuples may still be discovered.
			for (var columnIndex = 0; columnIndex < columns.length; columnIndex++) {
				var column = columns[columnIndex];
				var value = sourceData.getValue(rowIndex, columnIndex);
				if (!Column.isContextColumn(column)) {
					// Reading the value gets a bit odd; due to the aggregate none measures are actually reported
					// as tuples, for any column but the target column.
					var specObj = void 0;
					if (column.type !== 'measure') {
						specObj = value[0];
					}
					value = typeof specObj !== 'undefined' ? specObj.d : value.v;
					if (column.type === 'measure') {
						if (isMissingMeasureValue(value)) {
							if (Column.isTargetColumn(column)) {
								value = null; // normalize missing values to null
								includeRow = false;
							} else {
								value = colMean[columnIndex];
							}
						}
					} else {
						// column.type === 'categorical'
						if (isMissingCatValue(value)) {
							value = CAT_MISSING_VALUE;
						}
					}
				}
				// for context columns: use tuple index as value
				rowData[insertRowIndex * columns.length + columnIndex] = value;
			}
			if (includeRow) {
				// only count the row if it is included.
				insertRowIndex++;
			}
		}
		// In case any rows where filtered out: trim the row data to the correct length.
		rowData.length = insertRowIndex * columns.length;
		return new SourceTable(columns, rowData);
	}

	/**
  * SpecFormatter wraps Dashboard's formatSpec and complies wih VIPR's IFormatter interface
  */

	var SpecFormatter = function () {
		function SpecFormatter(_formatSpec) {
			_classCallCheck(this, SpecFormatter);

			Object.defineProperty(this, 'formatSpec', {
				enumerable: true,
				value: _formatSpec
			});
		}

		// IFormatter

		/**
   * Format a value. The type a formatter can format depends on the context.
   * @param {number} _value Value in a type the formatter can format.
   * @returns {string} The value, formatted as a String.
   */


		SpecFormatter.prototype.format = function format(_value) {
			return DashboardFormatter.format(_value, this.formatSpec);
		};

		return SpecFormatter;
	}();

	/**
  * SourceTable simply encodes the rows of a query result as tabular data.
  * It is used to create the machine learning model, and also as basis for DeltaTable to create a view that combines
  * the source data with predicted outputs.s
  */


	var SourceTable = function () {
		function SourceTable(_columns, _rowData) {
			_classCallCheck(this, SourceTable);

			Object.defineProperties(this, {
				columns: {
					enumerable: true,
					value: Object.freeze(_columns)
				},
				_rowData: {
					value: Object.freeze(_rowData)
				}
			});
		}

		/**
   * Get the column at the specified index.
   * Negative indices are taken from the end, e.g. index `-1` is the last column.
   * @param {number} _columnIndex The column index.
   * @returns {Column} The column definition.
   */
		SourceTable.prototype.getColumn = function getColumn(_columnIndex) {
			var index = Math.floor(_columnIndex) || 0;
			if (index < 0) {
				index += this.columnCount;
			}
			if (index < 0 || index >= this.columnCount) {
				throw new RangeError('_columnIndex ' + _columnIndex + ' is out of range');
			}
			return this.columns[index];
		};

		/**
   * Calculate the row offset in the 1 dimensional data array.
   * Accepts negative indices as "from end" and validates indices are in range.
   * @param {number} _rowIndex The row index.
   * @returns {number} The row offset in the `_rowData`.
   */


		SourceTable.prototype._rowOffset = function _rowOffset(_rowIndex) {
			var rowIndex = _rowIndex;
			if (rowIndex < 0) {
				rowIndex += this.rowCount;
			}
			if (rowIndex < 0 || rowIndex >= this.rowCount) {
				throw new RangeError('_rowIndex ' + _rowIndex + ' is out of range');
			}
			return rowIndex * this.columnCount;
		};

		/**
   * Get (part of) a row by index.
   * @param {number} _rowIndex The index of the row.
   * Negative values are interpreted as "from end", e.g. `-1` is the last row.
   * @param {number} _startColumn Optional: the first column to include. Defaults to `0`.
   * Negative values are interpreted as "from end", e.g. `-1` is the last column.
   * @param {number} _endColumn Optional: the end column to stop the row at. Defaults to `columnCount`.
   * Negative values are interpreted as "from end", e.g. `-1` is the last column.
   * @param {function} _transform Optional value transformation callback.
   * @returns {number[]} The values for the requested row (range).
   */


		SourceTable.prototype.getRow = function getRow(_rowIndex) {
			var _startColumn = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

			var _endColumn = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : this.columnCount;

			var _transform = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;

			var rowOffset = this._rowOffset(_rowIndex);
			var start = _startColumn;
			if (start < 0) {
				start += this.columnCount;
			}
			if (start < 0 || start > this.columnCount) throw new RangeError('_startColumn ' + _startColumn + ' is out of range');
			var end = _endColumn;
			if (end < 0) {
				end += this.columnCount;
			}
			if (end < 0 || end > this.columnCount) throw new RangeError('_endColumn ' + _endColumn + ' is out of range');
			if (end < start) end = start;
			var result = new Array(end - start);
			for (var i = start; i < end; i++) {
				var value = this._rowData[rowOffset + i];
				result[i - start] = _transform ? _transform(value, this.columns[i], i) : value;
			}
			return result;
		};

		/**
   * Get a value by row and column index.
   * @param {number} _rowIndex
   * @param {number} _columnIndex
   * @param {function|null} _transform Optional transformation on the value.
   * @returns {number}
   */


		SourceTable.prototype.getValue = function getValue(_rowIndex, _columnIndex) {
			var _transform = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;

			var rowOffset = this._rowOffset(_rowIndex);
			var columnIndex = _columnIndex;
			if (columnIndex < 0) columnIndex += this.columnCount;
			if (columnIndex < 0 || columnIndex >= this.columnCount) throw new RangeError('_columnIndex ' + _columnIndex + ' is out of range');
			var value = this._rowData[rowOffset + columnIndex];
			return _transform ? _transform(value, this.columns[columnIndex], columnIndex) : value;
		};

		/**
   * Get the row data as 1 dimensional Array.
   * @returns {number[]} The row data as 1 dimensional array.
   */


		SourceTable.prototype.getRowData = function getRowData() {
			return this._rowData;
		};

		/**
   * Determine how many columns have the specified kind.
   * @param {string} _kind The kind of the column, e.g. 'input', 'target' or 'context'.
   * @returns {number} The amount of columns that match the specified kind.
   */


		SourceTable.prototype.countColumnKind = function countColumnKind(_kind) {
			return this.columns.reduce(function (_acc, _column) {
				return _acc + (_column.kind === _kind ? 1 : 0);
			}, 0);
		};

		/**
   * Calculate the domain of a Column
   * @param {number|Column} _columnIndex The index of the column, or the column itself.
   * @returns {object} An object with min, max, sum and avg properties.
   */


		SourceTable.prototype.calculateDomain = function calculateDomain(_columnIndex) {
			var columnCount = this.columnCount;
			var rowCount = this.rowCount;
			var min = 0;
			var max = 0;
			var sum = 0;
			var columnIndex = void 0;
			if (Column.isColumn(_columnIndex)) {
				columnIndex = this.columns.indexOf(_columnIndex);
			} else {
				columnIndex = _columnIndex;
				if (columnIndex < 0) {
					columnIndex += columnCount;
				}
			}
			if (columnIndex < 0 || columnIndex >= columnCount) throw new RangeError('_columnIndex ' + _columnIndex + ' is out of range');
			for (var rowIndex = 0; rowIndex < rowCount; rowIndex++) {
				var value = this._rowData[rowIndex * columnCount + columnIndex];
				if (rowIndex === 0) {
					min = max = value;
				} else {
					if (min > value) {
						min = value;
					}
					if (max < value) {
						max = value;
					}
				}
				sum += value;
			}
			return {
				min: min,
				max: max,
				sum: sum,
				avg: rowCount > 0 ? sum / rowCount : 0
			};
		};

		_createClass(SourceTable, [{
			key: 'columnCount',
			get: function get() {
				return this.columns.length;
			}
		}, {
			key: 'rowCount',
			get: function get() {
				return this._rowData.length / this.columns.length;
			}
		}]);

		return SourceTable;
	}();

	// return the object that defines this module


	return {
		createTable: function createTable(_resultData, _slots) {
			var resultItemList = _resultData.getResultItemList();
			var columns = _resultData.getResultItemList().map(function (resultItem) {
				var mappingInfo = _slots.getMappingInfo(resultItem.getId());
				// Use the slot name as "kind". This will be 'context', 'input' or 'target'.
				var kind = mappingInfo.slot.getId();
				// No stacking support, so take the only column
				var sourceDataItem = resultItem.getDataItemList()[0];

				// possible values: 'integer' 'decimal', 'double', 'float', 'string', 'date', 'datetime', 'time'
				switch (sourceDataItem.getMetadataColumn().getDataType()) {
					case 'integer':
					case 'decimal':
					case 'double':
					case 'float':
						return new Column.MeasureColumn(sourceDataItem.getLabel(), sourceDataItem.getColumnId(), kind, sourceDataItem.getAggregation(), new SpecFormatter(mappingInfo.slot.getDefinition().getFormat()));

					// 'date', 'datetime', 'time' should not be features, but if they are, we consider them as categorical columns
					default:
						{
							// Odd API: the tuple count is reported via getRowCount().
							var tupleCount = resultItem.getRowCount();
							var tuples = new Array(tupleCount);
							for (var i = 0; i < tupleCount; i++) {
								var sourceTuple = resultItem.getValue(i);
								var sourceItem = sourceTuple[0];
								tuples[i] = new Decorated.Tuple(sourceItem.d, sourceItem.u);
							}
							return new Column.CategoricalColumn(sourceDataItem.getLabel(), sourceDataItem.getColumnId(), kind, Object.freeze(tuples));
						}
				}
			});

			// Process data into a new format: values that are either a tuple index or the actual value.
			var v1 = _resultData.getVersion() === '1';
			return v1 ? createSourceTable(_resultData.getDataPointValueList(), columns, resultItemList) : createSourceTableV2(_resultData, columns);
		}

	};
});
//# sourceMappingURL=SourceTable.js.map