'use strict';

/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM
 *| IBM Cognos Products: Smarts App
 *| (C) Copyright IBM Corp. 2020
 *|
 *| US Government Users Restricted Rights - Use, duplication or disclosure
 *| restricted by GSA ADP Schedule Contract with IBM Corp.
 *+------------------------------------------------------------------------+
 */
define(['jquery', 'dashboard-analytics/visualizations/renderer/VisView', './SourceTable', './DeltaTable', './Column', './KeyDrivers', './ScenarioUtils', './VidaUtils'], function ($, Base, SourceTable, DeltaTable, Column, KeyDrivers, ScenarioUtils, VidaUtils) {

	// defined messages to users
	var msg_title = 'What If Analysis';
	var msg_title_for = 'What If Analysis for {0}';
	var msg_create_model = 'Create model';
	var msg_train_model = 'Train model';
	var msg_training = 'Creating machine learning model. This process takes approximately 1 minute.';
	var msg_retraining = 'Applying additional training. This process takes approximately 1 minute.';
	var msg_autofill = 'Autofill input columns';
	var msg_filling = 'Please complete required fields in the fields tab.';
	var msg_few_points = 'Warning: specified columns do not contain sufficient data items. Please change your column selections or filters.';
	var msg_start_analysis = 'Click "Create model" to create a machine learning model and start your analysis.';
	var msg_model_prediction_strength = 'Prediction strength is {0}.';
	var msg_fetching_drivers = 'Autofilling Input Columns fields...';
	var msg_drivers_failed = 'Autofilling failed to complete. Please specify Input Columns manually.';
	var msg_drivers_succeeded = 'Found {0} input features';
	var msg_categorical = '(Categorical)';
	var msg_scenario_increase = '{0} +{1}';
	var msg_scenario_decrease = '{0} -{1}';
	var msg_and = ' and ';
	var msg_prediction_no_change = 'Prediction (no change)';

	var ATTR_INPUT_ID = 'data-input';

	var EPOCHS = 50;
	var TRAINING_ITERATIONS = 10;

	var MODEL_VERSION = 'v1';

	var StringEx = VidaUtils.getStringEx();

	var Widget = Base.extend({
		visModel: null,
		logger: null,
		ajaxSvc: null,
		whatIfServicePromise: null,
		/**
   * SourceTable instance containing the Dashboard query result.
   * This is used to create the training model, and as basis for the visualization data.
   */
		sourceTable: null,
		/**
   * DeltaTable contains a view to a predict result of the last executed scenario.
   * It can be processed into visualization data.
   */
		deltaTable: null,
		/**
   * ActiveScenario is the scenario that was added to represent the "live view" of changing the sliders.
   */
		activeScenario: null,
		/** Data for the Vida visualization. */
		viprData: null,
		/** whatif model. */
		model: null,
		/** whatif scenario. */
		scenario: null,
		/** whatif model prediction strength. */
		predictionStrength: null,
		/** whatif model metadata. */
		modelMetadata: null,
		/** whatif model metadata. */
		variableValues: null,
		/** allow restoring the scenario from dashboard if exists. */
		restoreScenario: true,
		/** Identifier of the target column. */
		targetColumnId: null,
		/** Aggregation of the target column. */
		targetColumnAggregation: null,
		/** Current prediction value. */
		prediction: null,
		config: null,
		/** The selected bundle id */
		_bundleId: null,
		/** A percentage forrmatter */
		_percentFmt: null,
		/** A percentage forrmatter */
		_numberFmt: null,

		// Nodes and Widgets
		_widgetTitleNode: null,
		_menuNode: null,
		_summaryNode: null,
		_trainButton: null,
		_inputButton: null,
		_widgetStatusNode: null,
		_inputsContainerNode: null,
		_inputsTBodyNode: null,
		_visualizationSectionNode: null,
		_vidaWidget: null,

		init: function init(_options) {
			Widget.inherited('init', this, arguments);
			this.content.setPropertyValue('titleMode', 'noTitle');

			Object.defineProperties(this, {
				visModel: {
					enumerable: true,
					value: _options.visModel
				},
				logger: {
					enumerable: true,
					value: _options.logger
				},
				ajaxSvc: {
					enumerable: true,
					value: _options.dashboardApi.getGlassCoreSvc('.Ajax')
				},
				whatIfServicePromise: {
					enumerable: true,
					value: _options.dashboardApi.getServiceWhenReady('.WhatIfService')
				},
				sourceTable: {
					enumerable: true,
					writable: true,
					value: null
				},
				deltaTable: {
					enumerable: true,
					writable: true,
					value: null
				},
				activeScenario: {
					enumerable: true,
					writable: true,
					value: null
				},
				viprData: {
					enumerable: true,
					writable: true,
					value: null
				},
				model: {
					enumerable: true,
					writable: true,
					value: null
				},
				modelMetadata: {
					enumerable: true,
					writable: true,
					value: null
				},
				variableValues: {
					enumerable: true,
					writable: true,
					value: null
				},
				predictionStrength: {
					enumerable: true,
					writable: true,
					value: null
				},
				restoreScenario: {
					enumerable: true,
					writable: true,
					value: true
				},
				targetColumnId: {
					enumerable: true,
					writable: true,
					value: null
				},
				targetColumnAggregation: {
					enumerable: true,
					writable: true,
					value: null
				},
				prediction: {
					enumerable: true,
					writable: true,
					value: null
				},
				config: {
					enumerable: true,
					writable: true,
					value: null
				},
				_bundleId: {
					writable: true,
					value: null
				},
				// TODO: get preferred locale via API call?
				_percentFmt: {
					value: ScenarioUtils.createPercentFormatter()
				},
				_numberFmt: {
					value: ScenarioUtils.createNumberFormatter()
				}
			});

			this._createWidgetLayout();
		},

		_hasColumnInSlot: function _hasColumnInSlot(inputSlot, colId) {
			var inputList = inputSlot.getDataItemList();
			for (var i = 0; i < inputList.length; i++) {
				if (inputList[i].getColumnId() === colId) {
					return true;
				}
			}
			return false;
		},

		_showVisualization: function _showVisualization(_visible) {
			this._visualizationSectionNode.css('display', _visible ? '' : 'none');
			this._summaryNode.css('display', _visible ? '' : 'none');
		},

		/**
   * Creates the general and static parts of the widget, such as the header, inputs area etc.
   */
		_createWidgetLayout: function _createWidgetLayout() {
			// The top section consists of the title, a "menu" and a status message
			var topSectionNode = $('<div></div>').appendTo(this.$el);

			var widgetTitleNode = $('<h4>' + msg_title + '</h4>').appendTo(topSectionNode);

			var menuNode = $('<div class="menu"></div>').appendTo(topSectionNode);

			// Train button
			var trainButton = $('<button class="btn btn-default">Create model</button>').appendTo(menuNode);
			trainButton.css({
				'margin-bottom': '5px',
				'margin-right': '10px',
				'height': '28px'
			});
			trainButton.click(this.train.bind(this));

			// Auto feature select button
			var inputButton = $('<button class="btn btn-default">' + msg_autofill + '</button>').appendTo(menuNode);
			inputButton.css({
				'margin-bottom': '5px',
				'margin-right': '10px',
				'height': '28px'
			});
			inputButton.click(this.autoFeatureSelect.bind(this));

			// Status
			var statusContainerNode = $('<span class="well well-sm"></span>').appendTo(menuNode);
			var widgetStatusNode = $('<span>' + msg_filling + '</span>').appendTo(statusContainerNode);

			var summaryNode = $('<table class="table table-striped table-bordered table-sm"></table>').appendTo(topSectionNode);
			summaryNode.css({ 'margin-top': '10px', 'margin-bottom': '10px' });
			var summaryTHeadNode = $('<thead></thead>').appendTo(summaryNode);
			var summaryBodyNode = $('<tbody></tbody>').appendTo(summaryNode);

			var tr = $('<tr/>').appendTo(summaryTHeadNode);
			//tr.css('padding-bottom', '20px');
			var th = $('<th/>').appendTo(tr);
			th.text('Baseline').css('text-align', 'center');

			th = $('<th/>').appendTo(tr);
			th.text('Predicted').css('text-align', 'center');

			th = $('<th/>').appendTo(tr);
			th.text('Variance').css('text-align', 'center');

			th = $('<th/>').appendTo(tr);
			th.text('% Variance').css('text-align', 'center');

			var headerHeight = '14em';

			// Inputs table (starts empty)
			var inputsContainerNode = $('<div style="position:absolute;left:0;top:' + headerHeight + ';width:45%;bottom:0;"></div>').appendTo(this.$el);
			var inputsTableNode = $('<table></table>').appendTo(inputsContainerNode);
			var inputsTBodyNode = $('<tbody></tbody>').appendTo(inputsTableNode);

			var visualizationSectionNode = $('<div style="position:absolute;top:' + headerHeight + ';right:0;width:50%;bottom:0;"></div>').appendTo(this.$el);
			// Visualization
			var vidaWidget = VidaUtils.getVidaWidget();

			visualizationSectionNode.append(vidaWidget.domNode);

			Object.defineProperties(this, {
				_widgetTitleNode: {
					value: widgetTitleNode
				},
				_menuNode: {
					value: menuNode
				},
				_summaryNode: {
					value: summaryNode
				},
				_summaryBodyNode: {
					value: summaryBodyNode
				},
				_summaryTHeadNode: {
					value: summaryTHeadNode
				},
				_trainButton: {
					value: trainButton
				},
				_inputButton: {
					value: inputButton
				},
				_widgetStatusNode: {
					value: widgetStatusNode
				},
				_inputsContainerNode: {
					value: inputsContainerNode
				},
				_inputsTBodyNode: {
					value: inputsTBodyNode
				},
				_visualizationSectionNode: {
					value: visualizationSectionNode
				},
				_vidaWidget: {
					value: vidaWidget
				},
				_actionHandler: {
					// The ActionHandler provides default interaction such as zoom, and provides callbacks for other
					// actions such as tooltip.
					value: VidaUtils.getActionHandler(vidaWidget)

				}
			});

			this._showVisualization(false);

			// input updates
			inputsTBodyNode.on('change', 'input[type="number"]', this._numberInputChange.bind(this));
			// sliders updates
			inputsTBodyNode.on('change', 'input[type="range"]', this._rangeInputChange.bind(this));
		},

		render: function render(renderInfo) {
			var _this = this;

			var renderVida = false;
			if (this.isMappingComplete()) {
				/*
     * Typically, users don not apply filters when training the model.
     * However, filters can be applied to the result (charts).
     * For example: in case the context slot shows a large number of values,
     * users should be able to filter the result, without re-training.
     * If data is refreshed, but the same columns presented (shape and order),
     * new SourceTable and DeltaTable are created.
     * */
				if (renderInfo.refresh.data || !this.sourceTable && renderInfo.data.getResult()) {
					var previousTable = this.sourceTable;
					this.sourceTable = SourceTable.createTable(renderInfo.data.getResult(), this.visualization.getSlots());
					var sameShape = !!previousTable && ScenarioUtils.sourceTableSameShape(this.sourceTable, previousTable);

					// Create a DeltaTable from the sourceTable
					this.deltaTable = DeltaTable.createDeltaTable(this.sourceTable);
					this.activeScenario = null;

					var contextDataItemCount = this.sourceTable.columns.reduce(function (_acc, _column) {
						return _acc + (Column.isContextColumn(_column) ? 1 : 0);
					}, 0);
					var inputDataItems = this.sourceTable.columns.filter(Column.isInputColumn);
					var targetDataItem = this.sourceTable.getColumn(-1);
					var rowCount = this.sourceTable.rowCount;
					var trainingData = {
						x: new Array(rowCount),
						y: new Array(rowCount)
					};
					// Populate training data
					for (var rowIndex = 0; rowIndex < rowCount; rowIndex++) {
						trainingData.x[rowIndex] = this.sourceTable.getRow(rowIndex, contextDataItemCount, -1, ScenarioUtils.valueForModel);
						trainingData.y[rowIndex] = this.sourceTable.getValue(rowIndex, -1, ScenarioUtils.valueForModel);
					}
					this.targetColumnId = targetDataItem.id;
					this.targetColumnAggregation = ScenarioUtils.toSupportedAggregation(targetDataItem.aggregation);

					if (!sameShape) {
						this.model = null;
						this.predictionStrength = null;
						this.viprData = null;
						renderVida = true;

						this._updateTrainButtonCaption();
						this.setStatus(rowCount < 100 ? msg_few_points : msg_start_analysis, 'danger');
						this.config = {
							epochs: EPOCHS,
							data: trainingData,
							types: inputDataItems.map(function (_column) {
								return _column.type;
							})
						};
						this.modelMetadata = {
							modelVersion: MODEL_VERSION,
							inputColTypes: this.config.types,
							inputColIds: inputDataItems.map(function (item) {
								return item.id;
							}),
							targetColId: targetDataItem.id,
							targetColType: targetDataItem.type
						};

						this._renderInputs(targetDataItem, inputDataItems);
						this.setEnabled(true);
						if (!previousTable && this.restoreScenario) {
							// this case when you re-open the dashboard
							this.restoreSavedScenario();
						} else {
							// clear model
							this.saveScenario();
						}
					} else {
						this.config.data = trainingData;
						this.predict();
					}
				} else if (this.viprData) {
					// If there is VIPR data, assume the visualization is visible and forward the render.
					renderVida = true;
				}
			} else {
				if (this.sourceTable) {
					this.sourceTable = null;
					this.deltaTable = null;
					this.viprData = null;
					this.activeScenario = null;
					this.config = null;
					this.model = null;
					// clear model
					this.saveScenario();
					this.targetColumnId = null;
					this.targetColumnAggregation = null;
					// Apply the "clear" state
					renderVida = true;

					this._updateTrainButtonCaption();
					// Clear the inputs table
					this._renderInputs(null, null);
				}
				// if mapping in not complete, we prevent restoring the model as a way
				// to start from scratch. We might add a restart button to deal with this.
				this.restoreScenario = false;
				this.setStatus(msg_filling, 'danger');
				this._setButtonsEnabled(false);
				var _targetSlotItems = this.visualization.getSlots().getSlot('target').getDataItemList();
				if (_targetSlotItems.length > 0) {
					this._inputButton.attr('disabled', null);
					// needed for autofilling rest call
					this.targetColumnId = _targetSlotItems[0].getColumnId();
					// needed for summary table
					this.targetColumnAggregation = ScenarioUtils.toSupportedAggregation(_targetSlotItems[0].aggregation);
				}
			}
			var targetSlotItems = this.visualization.getSlots().getSlot('target').getDataItemList();
			this.setWidgetTitle(targetSlotItems.length > 0 ? targetSlotItems[0].getLabel() : null);

			// Call super.render
			var result = Widget.inherited('render', this, arguments);
			if (renderVida) {
				if (!this.viprData) {
					this._vidaWidget.setData(null);
				}
				// Invoke Vida's render as soon as the widget render is complete.
				result = result.then(function () {
					return _this._vidaWidget.render('client');
				});
				// If there is no data, hide the visualization after render completes.
				if (!this.viprData) {
					// Sadly this is necessary since most Vida bundles do not clear on setData(null),
					// instead keeping the last state.
					result = result.then(function () {
						return _this._showVisualization(false);
					});
				}
			}
			return result;
		},

		/**
   * Renders the target and input fields in a table.
   * There are 3 columns, one for the label, one for the slider and one for the value.
   */
		_renderInputs: function _renderInputs(_targetDataItem, _inputColumns) {
			this._inputsTBodyNode.empty();
			var tr = void 0;
			var td = void 0;

			/*
   if (_targetDataItem) {
   	// Target
   	tr = $('<tr/>').appendTo(this._inputsTBodyNode);
   		td = $('<td colspan="3" style="padding:5px;"/>').appendTo(tr);
   	td.append('<h4>Target</h4>');
   		// show the target variable name (col 1)
   	tr = $('<tr/>').appendTo(this._inputsTBodyNode);
   	td = $('<td/>').appendTo(tr);
   	td.append('<i class="glyphicon status"></i> ' + _targetDataItem.label).css('padding', '2px').css('padding-left', '5px');
   		// no slider for target (col 2)
   	$('<td colspan="2"/>').appendTo(tr);
   }
   */

			if (_inputColumns) {
				// show the input variables header
				tr = $('<tr/>').appendTo(this._inputsTBodyNode);
				td = $('<td colspan="2" style="padding:10px;"/>').appendTo(tr);
				td = $('<td colspan="1" style="padding:10px;"/>').appendTo(tr);
				td.append('<h5>Input % change</h5>').css('text-align', 'center');

				// show the variables
				for (var i = 0; i < _inputColumns.length; i++) {
					// get data info
					var column = _inputColumns[i];
					tr = $('<tr/>').appendTo(this._inputsTBodyNode);

					// show the input variable name (col 1)
					td = $('<td/>').appendTo(tr);
					td.append('<i class="glyphicon glyphicon-transfer status"></i> ' + column.label).css('padding', '2px').css('padding-left', '5px');

					if (column.type === 'measure') {
						var input = void 0;
						// config values for sider
						// const domain = this.sourceTable.calculateDomain(column);

						td = $('<td/>').css('padding-left', '5px').css('padding-right', '10px').appendTo(tr);
						input = $('<input type="range" min="' + -100 + '" max="' + 100 + '" class="slider status">').appendTo(td);
						input.attr(ATTR_INPUT_ID, column.id);

						td = $('<td/>').css('padding-left', '5px').appendTo(tr);
						input = $('<input type="number" class="form-control status" style="padding:2px; height:22px" value="' + 0 + '">').appendTo(td);
						input.attr(ATTR_INPUT_ID, column.id);
					} else /*if(column.type === TYPE_CATEGORICAL)*/{
							/*
       // create all possible input options
       const options = column.tuples.map(function (_tuple, _index){
       	return '<option value=' + '"' + _index + '">' + _tuple.label + '</option>';
       });
       //Add no change option to the beginning of the array
       options.unshift('<option value="-1" style="font-style:italic;">' + msg_no_change + '</option>');
       	// drop down list
       td = $('<td colspan="2" style="padding: 10px 0"/>').appendTo(tr);
       let selector = $('<select id="value' + i + '" class="slider status" style="width:100%;" disabled/>').appendTo(td);
       selector.attr(ATTR_INPUT_ID, column.id);
       selector.append(options.join(''));
       */
							td = $('<td colspan="2" style="padding: 10px 0"/>').appendTo(tr);
							td.text(msg_categorical);
						}
				}
			}
		},

		predict: function predict() {
			var _this2 = this;

			if (!this.model) {
				// Cannot predict without a model
				return;
			}
			var inputColumns = this.sourceTable.columns.filter(Column.isInputColumn);
			/** The scenarioNameParts are added as a side effect of the variableValues map function */
			var scenarioNameParts = [];
			this.variableValues = inputColumns.map(function (_column) {
				// Simply persist categorical inputs by using the the source row value.
				if (_column.type === 'categorical') return null;
				// Build a selector to target the input field
				var attrSelector = '[' + ATTR_INPUT_ID + '="' + _column.id + '"]';
				var inputNode = this._inputsTBodyNode.find('input[type="number"]' + attrSelector + ',select' + attrSelector);
				// For measures, simply parse the value to a ratio.
				var delta = parseFloat(inputNode.val()) / 100;
				// Only include values that actually change.
				if (delta !== 0) {
					scenarioNameParts.push(StringEx.format(
					// Select the correct scenario message, increase or decrease
					delta < 0 ? msg_scenario_decrease : msg_scenario_increase,
					// Replacement 0: column name
					_column.label,
					// Replacement 1: the delta. Take the absolute, since we select an increase/decrease message
					this._percentFmt.format(Math.abs(delta))));
				}
				return delta;
			}, this);
			// New model inputs
			var newInputs = ScenarioUtils.changingColumnBy(this.config.data.x, inputColumns, this.variableValues);
			var newResults = this.model.predict(newInputs);
			// Create a scenario description from the deltas
			//const scenarioCaption = scenarioNameParts.length < 1 ? msg_prediction_no_change : StringEx.join(scenarioNameParts, ', ', msg_and);
			var scenarioCaption = 'Predicted';
			if (this.activeScenario) {
				this.activeScenario.update(scenarioCaption, newInputs, newResults);
			} else {
				this.activeScenario = this.deltaTable.addScenario(scenarioCaption, newInputs, newResults);
			}
			this._vidaWidget.newViz('bar').then(function (_bundle) {
				_this2._vidaWidget.setProperty('valueLabels.visible', true);
				_this2._mapDataToViz(_bundle.vizDef);
			});
			this.updateSummary(this.config.data.y, newResults);
			this.saveScenario();
		},
		restoreScenarioVariableValues: function restoreScenarioVariableValues() {
			var _this3 = this;

			if (this.variableValues) {
				var inputColumns = this.sourceTable.columns.filter(Column.isInputColumn);
				/** The scenarioNameParts are added as a side effect of the variableValues map function */
				var scenarioNameParts = [];
				inputColumns.forEach(function (_column, colIdx) {
					if (_column.type === 'measure') {
						var attrSelector = '[' + ATTR_INPUT_ID + '="' + _column.id + '"]';
						var inputNode = _this3._inputsTBodyNode.find('input[type="number"]' + attrSelector + ',select' + attrSelector);
						inputNode.val(_this3.variableValues[colIdx] * 100);
					}
				});
			}
		},
		/**
   * Create a slot mapping for the visualization and create data for it.
   * @param {VizDef} _vizDef The Vida visualization definition.
   */
		_mapDataToViz: function _mapDataToViz(_vizDef) {
			this.viprData = VidaUtils.mapDataToViz(_vizDef, this.deltaTable);
			// It may look odd that the data is passed twice; this first argument is the data, the second is the slot
			// mapping (which is also implemented by our viprData object).
			this._vidaWidget.setData( /*IVIPRData*/this.viprData, /*ISlotMapping*/this.viprData);
			this._showVisualization(true);
			return this._vidaWidget.render('client');
		},
		saveScenario: function saveScenario() {
			var _this4 = this;

			var whatIfFeatures = this.content.getFeature('WhatIfFeatures');

			if (this.model) {
				return this.whatIfServicePromise.then(function (_service) {
					if (_this4.model) {
						return _service.serializeWhatIfModel(_this4.model);
					}
				}).then(function (_response) {

					var scenario = {
						model: _response,
						modelMetadata: _this4.modelMetadata,
						modelMetrics: _this4.modelMetrics,
						variableValues: _this4.variableValues,
						predictionStrength: _this4.predictionStrength
					};
					_this4.restoreScenario = true;
					whatIfFeatures.setScenario(scenario);
				});
			} else {
				// clear model
				this.restoreScenario = false;
				whatIfFeatures.setScenario(null);
			}
		},

		restoreSavedScenario: function restoreSavedScenario() {
			var _this5 = this;

			var whatIfFeatures = this.content.getFeature('WhatIfFeatures');
			var scenario = whatIfFeatures.toJSON();
			if (scenario && scenario.model && scenario.modelMetadata && ScenarioUtils.sameModelMetadata(scenario.modelMetadata, this.modelMetadata)) {
				return this.whatIfServicePromise.then(function (_service) {
					return _service.buildWhatIfModelFromJson(scenario.model);
				}).then(function (_response) {
					_this5.model = _response;
					_this5.modelMetrics = scenario.modelMetrics;
					_this5.predictionStrength = scenario.predictionStrength;
					/*
      * evaluate model on the new data,
      * this data might be different than
      * the data fetched when trained the model,
      * data in data sources could be updated.
      *
      * TODO: if model performance is significantly different
      * than what is reported when the model saved, notify the user
      * compare modelMetrics with this.modelMetrics
     */
					//const modelMetrics = this.model.evaluate(this.config.data.x,this.config.data.y);
					_this5.setStatus(StringEx.format(msg_model_prediction_strength, _this5.predictionStrength), 'success');
					_this5.setEnabled(true);
					// continue with a prediction
					if (scenario.variableValues) {
						_this5.variableValues = scenario.variableValues;
						_this5.restoreScenarioVariableValues();
					}
					_this5.predict();
				}).catch(function (_error) {
					_this5.logger.error(_error);
					_this5.setError('Something went wrong while restoring the model');
				});
			}
		},

		train: function train() {
			return this.recursiveTraining(TRAINING_ITERATIONS);
		},
		recursiveTraining: function recursiveTraining(iterationNo) {
			var _this6 = this;

			this.setEnabled(false);
			return this.whatIfServicePromise.then(function (_service) {
				// Either continue training, or create a new model
				if (_this6.model) {
					if (TRAINING_ITERATIONS === iterationNo) {
						_this6.setStatus(msg_retraining, 'danger');
					}
					return _service.continueTraining(_this6.model, _this6.config.data);
				} else {
					if (TRAINING_ITERATIONS === iterationNo) {
						_this6.setStatus(msg_training, 'danger');
					}
					return _service.createModel(_this6.config);
				}
			}).then(function (_response) {
				_this6.model = _response.model;
				_this6._updateTrainButtonCaption();
				var predictionStrength = 1 - Math.sqrt(1 - Math.max(0, _response.r2));
				// TODO: add ', Standard Error is ±' + _response.mae.toFixed(2)'
				_this6.predictionStrength = _this6._percentFmt.format(predictionStrength);
				var completePercent = Math.round((TRAINING_ITERATIONS + 1 - iterationNo) / (TRAINING_ITERATIONS + 1) * 100);
				_this6.setStatus(StringEx.format(msg_model_prediction_strength + ' Training progress: ' + completePercent + '%.', _this6.predictionStrength), 'success');
				if (iterationNo <= 0 || _response.stopTraining) {
					// this will evaluate the model over all data, where _response.r2 is computed over
					// a validation set that the model never seen while training. This is needed when
					// storing the model to check if the model performance
					// is similar to the trained model.
					//this.modelMetrics = this.model.evaluate(this.config.data.x, this.config.data.y);
					_this6.setEnabled(true);
					_this6.setStatus(StringEx.format(msg_model_prediction_strength, _this6.predictionStrength), 'success');
					// continue with a prediction
					_this6.predict();
				} else {
					return _this6.recursiveTraining(iterationNo - 1);
				}
			}).catch(function (_error) {
				_this6.logger.error(_error);
				_this6.setError('Something went wrong while training');
			});
		},
		autoFeatureSelect: function autoFeatureSelect() {
			var _this7 = this;

			this.setStatus(msg_fetching_drivers, 'danger');

			KeyDrivers.getKeyDrivers(this.ajaxSvc, this.visModel.getModule(), this.targetColumnId).then(function (drivers) {
				var inputSlot = _this7.visualization.getSlots().getSlot('input');
				var items = [];

				for (var i = 0; i < drivers.length; i++) {
					if (!_this7._hasColumnInSlot(inputSlot, drivers[i])) {
						items.push({
							columnId: drivers[i]
						});
					}
				}
				_this7.setStatus(StringEx.format(msg_drivers_succeeded, items.length), 'success');
				inputSlot.addDataItems(items);
			}).catch(function (_error) {
				_this7.logger.error(_error);
				_this7.setStatus(msg_drivers_failed, 'danger');
			});
		},

		_numberInputChange: function _numberInputChange(_event) {
			var node = _event.target;
			this.$el.find('input[type="range"][' + ATTR_INPUT_ID + '="' + node.getAttribute(ATTR_INPUT_ID) + '"]').val(node.value);
			this.predict();
		},

		_rangeInputChange: function _rangeInputChange(_event) {
			var node = _event.target;
			this.$el.find('input[type="number"][' + ATTR_INPUT_ID + '="' + node.getAttribute(ATTR_INPUT_ID) + '"]').val(node.valueAsNumber);
			this.predict();
		},

		setWidgetTitle: function setWidgetTitle(_target) {
			this._widgetTitleNode.text(_target ? StringEx.format(msg_title_for, _target) : msg_title);
		},

		updateSummary: function updateSummary(_baselineArr, _predictedArr) {
			this._summaryBodyNode.empty();
			var _baseline = _baselineArr.reduce(function (a, b) {
				return a + b;
			}, 0);
			var _predicted = _predictedArr.reduce(function (a, b) {
				return a + b;
			}, 0);
			if (this.targetColumnAggregation === 'avg' && _baselineArr.length > 0 && _predictedArr.length > 0) {
				_baseline = _baseline / _baselineArr.length;
				_predicted = _predicted / _predictedArr.length;
			}
			var tr = $('<tr/>').appendTo(this._summaryBodyNode);
			// show the input variable name (col 1)
			var td = $('<td/>').appendTo(tr);
			td.text(this._numberFmt.format(_baseline)).css('text-align', 'center');

			td = $('<td/>').appendTo(tr);
			td.text(this._numberFmt.format(_predicted)).css('text-align', 'center');

			td = $('<td/>').appendTo(tr);
			td.text(this._numberFmt.format(_predicted - _baseline)).css('text-align', 'center');

			td = $('<td/>').appendTo(tr);
			var percent = (_predicted - _baseline) / _baseline;
			if (_baseline == _predicted) {
				// this is to handle case when both _baseline and _predicted equal to zero
				percent = 1;
			} else if (_baseline == 0) {
				// this is to handle case when _baseline ==0
				percent = (_predicted - 1e-10) / 1e-10;
			}
			td.text(this._percentFmt.format(percent)).css('text-align', 'center');
		},
		setStatus: function setStatus(_message, _style) {
			this._widgetStatusNode.text(_message).toggleClass('text-muted', _style === 'muted').toggleClass('text-danger', _style === 'danger').toggleClass('text-success', _style === 'success');
		},

		_updateTrainButtonCaption: function _updateTrainButtonCaption() {
			this._trainButton.text(this.model ? msg_train_model : msg_create_model);
		},

		_setInputsEnabled: function _setInputsEnabled(_enabled) {
			var inputs = this._inputsTBodyNode.find('select[' + ATTR_INPUT_ID + '],input[' + ATTR_INPUT_ID + ']');
			inputs.attr('disabled', _enabled ? null : 'disabled');
		},

		_setButtonsEnabled: function _setButtonsEnabled(_enabled) {
			var inputs = this._menuNode.find('button');
			inputs.attr('disabled', _enabled ? null : 'disabled');
		},

		setEnabled: function setEnabled(_enabled) {
			this._setButtonsEnabled(_enabled);
			this._setInputsEnabled(_enabled && !!this.model);
		},

		setError: function setError(_message) {
			this.setStatus(_message, 'danger');
			this.setEnabled(true);
		}

	});

	return Widget;
});
//# sourceMappingURL=what_if.js.map